"""server.py"""

import json
import pkgutil
import logging

from requests import Request, Session
from requests.exceptions import InvalidSchema, RequestException
import jsonschema

from jsonrpcclient import rpc
from jsonrpcclient import exceptions


logger = logging.getLogger(__name__)
request_log = logging.getLogger(__name__+'.request')
response_log = logging.getLogger(__name__+'.response')

DEFAULT_HTTP_HEADERS = {
    'Content-Type': 'application/json',
    'Accept': 'application/json'
}

class Server(object):
    """This class acts as the remote server"""

    def __init__(self, endpoint, **kwargs):
        """Instantiate a remote server object.
        >>> server = Server('http://example.com/api', \
                headers={'Content-Type': 'application/json-rpc'}, \
                auth=('user', 'pass'))
        """
        kwargs.setdefault('headers', DEFAULT_HTTP_HEADERS)
        self.endpoint = endpoint
        self.headers = kwargs['headers']
        self.requests_kwargs = kwargs
        kwargs.pop('headers')

    def __getattr__(self, name):
        """Catch undefined methods and handle them as RPC requests.
        The technique is here: http://code.activestate.com/recipes/307618/
        """
        def attr_handler(*args, **kwargs):
            """Call self.request from here"""
            if kwargs.get('response', False):
                return self.request(name, *args, **kwargs)
            else:
                return self.notify(name, *args, **kwargs)
        return attr_handler

    def request(self, method_name, *args, **kwargs):
        """JSON-RPC Request (expect a response)"""
        kwargs['response'] = True
        return self.handle_response(
            self.send_message(rpc.request(method_name, *args, **kwargs)), True)

    def notify(self, method_name, *args, **kwargs):
        """JSON-RPC Notification (no response required)"""
        return self.handle_response(
            self.send_message(rpc.request(method_name, *args, **kwargs)), False)

    def send_message(self, request):
        """Send the RPC request (a json dict) to the server.
        Calls a procedure on another server.
        Raises JsonRpcClientError: On any error caught.
        """
        logger.debug('Sending via http post...')
        s = Session()
        # Prepare the request
        request = Request(method='POST', url=self.endpoint, \
            headers=self.headers, json=request, **self.requests_kwargs)
        request = s.prepare_request(request)
        request.headers = dict(list(dict(request.headers).items()) + list(
            self.headers.items()))
        # Log the request before sending
        request_log.info(
            request.body,
            extra={
                'http_headers': request.headers
            })
        try:
            response = s.send(request)
        # Catch the requests module's InvalidSchema exception if the json is
        # invalid.
        except InvalidSchema:
            raise exceptions.InvalidRequest()
        # Catch all other requests exceptions, such as network issues.
        # See http://stackoverflow.com/questions/16511337/
        except RequestException: # Base requests exception
            raise exceptions.ConnectionError()
        finally:
            s.close()
        # Log the response, cleaning it up a bit
        response_log.info(
            response.text \
                .replace("\n", '').replace('  ', ' ').replace('{ ', '{'),
            extra={
                'http_code': response.status_code,
                'http_reason': response.reason,
                'http_headers': response.headers
            })
        # Raise exception if the HTTP status code was not 200-299.
        if response.status_code not in range(200, 300):
            raise exceptions.Non2xxResponse(response.status_code)
        return response.text

    @staticmethod
    def handle_response(response, expected_response=False):
        """Processes the response (a json string)"""
        # A response was expected, but none was given?
        if expected_response and not len(response):
            raise exceptions.ReceivedNoResponse()
        # Was a response given?
        if len(response):
            # Attempt to parse the response
            try:
                response_dict = json.loads(response)
            except ValueError:
                raise exceptions.ParseResponseError()
            # Unwanted response - A response was not asked for, but one was
            # given anyway. It may not be necessary to raise here.
            if not expected_response and 'result' in response_dict:
                raise exceptions.UnwantedResponse()
            # Validate the response against the Response schema
            try:
                jsonschema.validate(response_dict, json.loads(pkgutil.get_data(
                    __name__, 'response-schema.json').decode('utf-8')))
            except jsonschema.ValidationError:
                raise exceptions.InvalidResponse()
            # If the response was "error", raise it, to ensure it's handled
            if 'error' in response_dict:
                raise exceptions.ReceivedErrorResponse(
                    response_dict['error']['code'],
                    response_dict['error']['message'])
            # Otherwise, surely we have a result to return
            return response_dict['result']

        return None



"""exceptions.py"""


class JsonRpcClientError(Exception):
    """Base class for the other exceptions"""
    pass


class InvalidRequest(JsonRpcClientError):
    """The request being sent is not valid JSON."""
    def __init__(self):
        super(InvalidRequest, self).__init__(
            'The request you\'re sending is not valid json')


class ConnectionError(JsonRpcClientError): # pylint: disable=redefined-builtin
    """There was a network issue, invalid HTTP response or timeout."""
    def __init__(self):
        super(ConnectionError, self).__init__(
            'Connection error')


class ReceivedNoResponse(JsonRpcClientError):
    """A response message was expected, but none was given."""
    def __init__(self):
        super(ReceivedNoResponse, self).__init__(
            'No response was received')


class Non2xxResponse(JsonRpcClientError):
    """The server responded with a HTTP status code other than 200-299."""
    def __init__(self, status_code):
        super(Non2xxResponse, self).__init__(
            'Returned status code '+str(status_code))


class UnwantedResponse(JsonRpcClientError):
    """A response was not requested, but one was given anyway."""
    def __init__(self):
        super(UnwantedResponse, self).__init__(
            'An unwanted response was given')


class ParseResponseError(JsonRpcClientError):
    """The response was not valid json."""
    def __init__(self):
        super(ParseResponseError, self).__init__(
            'The response was not valid json')


class InvalidResponse(JsonRpcClientError):
    """The response was not a valid JSON-RPC response."""
    def __init__(self):
        super(InvalidResponse, self).__init__(
            'The response was not a valid json-rpc 2.0 response')


class ReceivedErrorResponse(JsonRpcClientError):
    """The server gave a valid JSON-RPC *error* response."""
    def __init__(self, code, message): #pylint:disable=unused-argument
        super(ReceivedErrorResponse, self).__init__(message)



"""rpc.py"""

import itertools
import json #pylint:disable=unused-import
from collections import OrderedDict

id_generator = itertools.count(1) # First generated is 1


def request(method, *args, **kwargs):
    #pylint:disable=line-too-long
    """Returns a JSON-RPC 2.0 request, in OrderedDict format. Convert to a json
    string with json.dumps().

    Notification
    >>> json.dumps(request('go'))
    '{"jsonrpc": "2.0", "method": "go"}'

    Passing both positional and keyword arguments
    >>> json.dumps(request('find', 'Foo', age=42))
    '{"jsonrpc": "2.0", "method": "find", "params": ["Foo", {"age": 42}]}'

    Requests (requiring a response)
    >>> json.dumps(request('add', 2, 3, response=True))
    '{"jsonrpc": "2.0", "method": "add", "params": [2, 3], "id": 1}'
    """
    # Get the request id
    request_id = None
    if kwargs.get('response', False):
        request_id = next(id_generator)
    # Pop 'response' out of the kwargs if present
    if 'response' in kwargs:
        kwargs.pop('response')
    # jsonrpc, method
    r = OrderedDict([
        ('jsonrpc', '2.0'),
        ('method', method)
    ])
    # Get the params
    params = list()
    if args:
        for i in args:
            params.append(i)
    if kwargs:
        params.append(OrderedDict(sorted(kwargs.items())))
    if len(params):
        # If there's only param and it's a dict or list, take it out of the
        # params list, rather than having a list within a list [[]]
        if len(params) == 1 and (
                isinstance(params[0], dict) or isinstance(params[0], list)):
            params = params[0]
        r.update(OrderedDict([
            ('params', params)
        ]))
    # response_id
    if request_id:
        r.update(OrderedDict([
            ('id', request_id)
        ]))
    return r



"""setup.py"""
#pylint:disable=line-too-long

import sys
from setuptools.command.test import test as TestCommand
from codecs import open as codecs_open

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup #pylint:disable=import-error,no-name-in-module

with codecs_open('README.rst', 'r', 'utf-8') as f:
    readme = f.read()

with codecs_open('HISTORY.rst', 'r', 'utf-8') as f:
    history = f.read()

class Tox(TestCommand):
    user_options = [('tox-args=', 'a', "Arguments to pass to tox")]
    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = '-v'
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import tox
        import shlex
        errno = tox.cmdline(args=shlex.split(self.tox_args))
        sys.exit(errno)

setup(
    name='jsonrpcclient',
    version='1.1.4',
    description='JSON-RPC client library.',
    long_description=readme + '\n\n' + history,
    author='Beau Barker',
    author_email='beauinmelbourne@gmail.com',
    url='https://jsonrpcclient.readthedocs.org/',
    packages=['jsonrpcclient'],
    package_data={'jsonrpcclient': ['response-schema.json']},
    include_package_data=True,
    install_requires=['requests', 'jsonschema'],
    tests_require=['tox'],
    cmdclass = {'test': Tox},
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: OS Independent'
        ],
    )
